@@ -1,6 +1,6 @@
 use 5.020;
 use warnings;
-package Plate 0.3;
+package Plate 0.4;
 
 use Carp 'croak';
 use File::Spec;
@@ -42,8 +42,9 @@ my $re_pre = qr'(.*?)(?:
     ^%%\h*(\V*?)\h*(?:\R|\z)|
     ^<%%(perl)>(?:\R|\z)(?:(.*?)^</%%\g-2>(?:\R|\z))?|
     ^<%%def\h+([\w/\.-]+)>(?:\R|\z)|
-    <%%\h*(.+?)\h*(?:\|\h*(|\w+(?:\h*\|\h*\w+)*)\h*)?%%>|
-    <&&(\|)?\h*(.+?)\h*(?:\|\h*(|\w+(?:\h*\|\h*\w+)*)\h*)?&&>|
+    <%%(\s*(?:\#.+?)?)%%>|
+    <%%\s*(.+?)\s*(?:\|\h*(|\w+(?:\s*\|\h*\w+)*)\s*)?%%>|
+    <&&(\|)?\s*(.+?)\s*(?:\|\h*(|\w+(?:\s*\|\h*\w+)*)\s*)?&&>|
     </(%%def|%%perl)>(?:\R|\z)|
     </(&&)>|\z
 )'mosx;
@@ -51,15 +52,16 @@ my $re_run = qr'(.*?)(?:
     ^%\h*(\V*?)\h*(?:\R|\z)|
     ^<%(perl)>(?:\R|\z)(?:(.*?)^</%\g-2>(?:\R|\z))?|
     ^<%def\h+([\w/\.-]+)>(?:\R|\z)|
-    <%\h*(.+?)\h*(?:\|\h*(|\w+(?:\h*\|\h*\w+)*)\h*)?%>|
-    <&(\|)?\h*(.+?)\h*(?:\|\h*(|\w+(?:\h*\|\h*\w+)*)\h*)?&>|
+    <%(\s*(?:\#.+?)?)%>|
+    <%\s*(.+?)\s*(?:\|\h*(|\w+(?:\s*\|\h*\w+)*)\s*)?%>|
+    <&(\|)?\s*(.+?)\s*(?:\|\h*(|\w+(?:\s*\|\h*\w+)*)\s*)?&>|
     </(%def|%perl)>(?:\R|\z)|
     </(&)>|\z
 )'mosx;
 
 sub _parse_text {
     my $text = $_[0];
-    $_[2] = $text =~ s/\\\R//g if $_[1] == $re_run;
+    $_[2] = $text =~ s/\\\R//g unless $_[1];
     $text =~ s/(\\|')/\\$1/g;
     length $text ? "'$text'" : ();
 }
@@ -93,10 +95,12 @@ sub _parse_fltr {
 sub _parse {
     my @expr;
     my $stmt;
+    my $pre = $_[1] == $re_pre;
     my $fix_line_num;
     my $expr2stmt = sub {
         if (@expr) {
             if (defined $stmt) {
+                $stmt .= ';push@Plate::_l,length$Plate::_b,__LINE__' if $pre;
                 $stmt .= ';$Plate::_b.=';
             } else {
                 $stmt = 'local$Plate::_b=';
@@ -110,12 +114,12 @@ sub _parse {
     while ($_[0] =~ /$_[1]/g) {
 
         if (length $1) {
-            push @expr, _parse_text $1, $_[1], my $add_lines;
+            push @expr, _parse_text $1, $pre, my $add_lines;
             (@expr ? $expr[-1] : defined $stmt ? $stmt : ($expr[0] = "''")) .= "\n" x $add_lines if $add_lines;
             $fix_line_num = @expr if $fix_line_num;
         }
 
-        if ($_[1] == $re_run and @Plate::_l and $Plate::_l[0] <= $+[1]) {
+        if (!$pre and @Plate::_l and $Plate::_l[0] <= $+[1]) {
             my($pos, $line) = splice @Plate::_l, 0, 2;
             ($pos, $line) = splice @Plate::_l, 0, 2 while @Plate::_l and $Plate::_l[0] <= $+[1];
             my $rem = $+[1] - $pos;
@@ -135,8 +139,8 @@ sub _parse {
             $expr2stmt->();
             unless (defined $4) {
                 my $line = 1 + $stmt =~ y/\n//;
-                $line = "$_[2] line $line.\nPlate ".($_[1] == $re_pre && 'pre').'compilation failed';
-                my $tag = ($_[1] == $re_pre && '%').'%'.$3;
+                $line = "$_[2] line $line.\nPlate ".($pre && 'pre').'compilation failed';
+                my $tag = ($pre && '%').'%'.$3;
                 croak "Opening <$tag...> tag without closing </$tag> tag at $line";
             }
             $stmt .= "\n$4\n";
@@ -144,61 +148,72 @@ sub _parse {
         } elsif (defined $5) {
             # <%def ...>
             $expr2stmt->();
-            local $_[3] = ($_[1] == $re_pre && '%').'%def';
+            local $_[3] = ($pre && '%').'%def';
             $stmt .= 'local$$Plate::_s{mem}{'._parse_defn($5)."}=\nsub{".&_parse.'};';
 
         } elsif (defined $6) {
+            # <%# ... %>
+            my $add_lines = $6 =~ tr/\n//;
+            (@expr ? $expr[-1] : defined $stmt ? $stmt : ($expr[0] = "''")) .= "\n" x $add_lines if $add_lines;
+            $fix_line_num = @expr if $fix_line_num;
+
+        } elsif (defined $7) {
             # <% ... %>
+            my $nl1 = "\n" x substr($_[0], $+[1], $-[7] - $+[1]) =~ tr/\n//;
+            my $nl2 = "\n" x substr($_[0], $+[7], $+[0] - $+[7]) =~ tr/\n//;
             $expr2stmt->(1) if $fix_line_num;
             $fix_line_num = push @expr,
-            _parse_fltr "do{$6}", $7 // $$Plate::_s{auto_filter};
+            _parse_fltr "do{$nl1$7}$nl2", $8 // $$Plate::_s{auto_filter};
+            $expr2stmt->() if $pre;
 
-        } elsif (defined $9) {
+        } elsif (defined $10) {
             # <& ... &> or <&| ... &>
-            my($tmpl, $args) = do { $9 =~ /^([\w\/\.-]+)\s*(?:,\s*(.*))?$/ };
-            $expr2stmt->(1) if $fix_line_num;
+            my $nl = "\n" x (substr($_[0], $+[1], $+[0] - $+[1]) =~ tr/\n// - $10 =~ tr/\n//);
+            my($tmpl, $args) = do { $10 =~ /^([\w\/\.-]+)\s*(?:,\s*(.*))?$/s };
+            $expr2stmt->(!$pre) if $pre or $fix_line_num;
             if (defined $tmpl) {
                 if ($tmpl eq '_') {
-                    $fix_line_num = push @expr, _parse_fltr defined $8
+                    $fix_line_num = push @expr, _parse_fltr defined $9
                     ? do {
                         $args = defined $args ? "($args)" : '';
-                        local $_[3] = $_[1] == $re_pre ? '&&' : '&';
-                        '(@Plate::_c?do{local@Plate::_c=@Plate::_c;&{splice@Plate::_c,-1,1,sub{'.&_parse."}}$args}:undef)"
+                        local $_[3] = $pre ? '&&' : '&';
+                        '(@Plate::_c?do{local@Plate::_c=@Plate::_c;&{splice@Plate::_c,-1,1,sub{'.&_parse."}}$args}:undef)$nl"
                     }
-                    : defined $args ? "Plate::content($args)" : '&Plate::content', $10;
+                    : defined $args ? "do{Plate::content($args)}$nl" : "do{&Plate::content}$nl", $11;
+                    $expr2stmt->() if $pre and $nl;
                     next;
                 }
                 $tmpl = defined $args ? "Plate::_r('$tmpl',($args)," : "Plate::_r('$tmpl',";
             } else {
-                $tmpl = "Plate::_r($9,";
+                $tmpl = "Plate::_r($10,";
             }
             $fix_line_num = push @expr,
-            _parse_fltr $tmpl.(defined $8 ? (local $_[3] = $_[1] == $re_pre ? '&&' : '&', 'sub{'.&_parse.'}') : 'undef').')', $10;
+            _parse_fltr "do{$tmpl".(defined $9 ? (local $_[3] = $pre ? '&&' : '&', 'sub{'.&_parse.'}') : 'undef').")}$nl", $11;
+            $expr2stmt->() if $pre and $nl;
 
         } else {
             # </%...> or </&> or \z
-            my $tag = $11 // $12 // '';
+            my $tag = $12 // $13 // '';
             if ($tag ne $_[3]) {
                 my $line = 1 + join('', $stmt // '', @expr) =~ y/\n//;
-                $line = "$_[2] line $line.\nPlate ".($_[1] == $re_pre && 'pre').'compilation failed';
+                $line = "$_[2] line $line.\nPlate ".($pre && 'pre').'compilation failed';
                 croak $tag
                 ? "Closing </$tag> tag without opening <$tag...> tag at $line"
                 : "Opening <$_[3]...> tag without closing </$_[3]> tag at $line";
             }
 
             my $pl = defined $stmt
-            ? $stmt.join('.', ';$Plate::_b', @expr)
+            ? do {
+                $stmt .= ';push@Plate::_l,length$Plate::_b,__LINE__' if $pre and @expr;
+                $stmt.join('.', ';$Plate::_b', @expr);
+            }
             : @expr ? join('.', @expr) : "''";
-            $pl .= '=~s/\R\z//r' if $_[1] == $re_run and $$Plate::_s{chomp};
-            $pl .= "\n" if defined $11;
+            $pl .= '=~s/\R\z//r' if !$pre and $$Plate::_s{chomp};
+            $pl .= "\n" if defined $12;
             return $pl;
         }
 
-        if ($_[1] == $re_pre) {
-            $expr2stmt->();
-            $stmt .= 'push@Plate::_l,length$Plate::_b,__LINE__;';
-
-        } elsif (@Plate::_l and $Plate::_l[0] <= $+[0]) {
+        if (!$pre and @Plate::_l and $Plate::_l[0] <= $+[0]) {
             my($pos, $line) = splice @Plate::_l, 0, 2;
             ($pos, $line) = splice @Plate::_l, 0, 2 while @Plate::_l and $Plate::_l[0] <= $+[0];
             my $rem = $+[0] - $pos;
@@ -418,6 +433,15 @@ To explicitly avoid the default filter use the empty string as a filter.
 
 Lines that start with a C<%> character are treated as Perl statements.
 
+=head3 Comments
+
+    %# Comment line
+    <% # inline comment %>
+    <%#
+        Multi-line
+        comment
+    %>
+
 =head3 Perl blocks
 
     <%perl>
@@ -515,6 +539,16 @@ the final newline in every template will be removed.
 
 Set this to the encoding of your template files.
 
+=item C<< filters => { html => \&_basic_html_filter } >>
+
+A hash of filters to set for use in templates.
+The key is the name of the filter, and the value is the CODE ref, subroutine name or C<undef>.
+The subroutine will be given one argument (the content to filter) as a string,
+and must return the filtered string.
+To remove a filter pass C<undef> as it's value.
+
+To remove all filters pass C<undef> instead of a HASH ref.
+
 =item C<< keep_undef => undef >>
 
 If set to a false value (the default),
@@ -552,6 +586,17 @@ The suffix appended to template names when searching on the filesystem.
 
 The C<umask> used when creating cache files and directories.
 
+=item C<< vars => {} >>
+
+A hash of vars to set for use in templates.
+This will define new local variables to be imported into the templating package when compiling and running templates.
+If the value is not a reference it will be a constant in the templating package.
+To remove a var pass C<undef> as it's value.
+
+To remove all vars pass C<undef> instead of a HASH ref.
+
+All templates will have access to these variables, subroutines and constants even under C<use strict>.
+
 =back
 
 =cut
@@ -638,60 +683,6 @@ sub has_content {
     @Plate::_c and $Plate::_c[-1] != \&_empty;
 }
 
-=head2 filter
-
-    $plate->filter($filter_name => sub { ... });
-
-Add a new filter for use in templates.
-The subroutine will be given one argument (the content to filter) as a string,
-and must return the filtered string.
-
-=cut
-
-sub filter {
-    my($self, $name, $code) = @_;
-
-    $name =~ /^\w+$/
-        or croak "Invalid filter name '$name'";
-    return delete $$self{filters}{$name} unless defined $code;
-    ref $code eq 'CODE'
-        or $code = ($code =~ /(.*)::(.*)/ ? $1->can($2) : do { my($i,$p) = 0; $i++ while __PACKAGE__ eq ($p = caller $i); $p->can($code) })
-        or croak "Invalid subroutine '$_[2]' for filter '$name'";
-    $$self{filters}{$name} = $code;
-}
-
-=head2 var
-
-    $plate->var('$var' => \$var);
-    $plate->var('%hash' => \%hash);
-    $plate->var('@array' => \@array);
-    $plate->var(func => \&func);
-    $plate->var(CONST => 123);
-
-Define a new local variable to be imported into the templating package when compiling and running templates.
-If the value is not a reference it will be a constant in the templating package.
-To remove a var pass C<undef> as the value.
-
-All templates will have access to these variables, subroutines and constants even under C<use strict>.
-
-=cut
-
-my %sigil = (
-    ARRAY => '@',
-    CODE => '&',
-    GLOB => '*',
-    HASH => '%',
-);
-sub var {
-    my($self, $name, $ref) = @_;
-
-    defined $ref or return delete $$self{vars}{$name};
-
-    my $sigil = $sigil{Scalar::Util::reftype $ref // 'CODE'} // '$';
-    $name =~ s/^\Q$sigil\E?/$sigil ne '&' && $sigil/e;
-    $$self{vars}{$name} = $ref;
-}
-
 =head2 define
 
     $plate->define($template_name => $content);
@@ -766,6 +757,13 @@ Options are the same as those for L</new>.
 
 =cut
 
+my %sigil = (
+    ARRAY => '@',
+    CODE => '&',
+    GLOB => '*',
+    HASH => '%',
+);
+
 sub set {
     my($self, %opt) = @_;
 
@@ -780,11 +778,43 @@ sub set {
             $v = _path $v, 1 if defined $v;
         } elsif ($k =~ /^(?:(?:cache_)?suffix|init|io_layers|once)$/) {
             $v //= '';
-        } elsif ($k eq 'filters' or $k eq 'vars') {
-            my $method = substr $k, 0, -1;
+        } elsif ($k eq 'filters') {
             if (defined $v) {
                 ref $v eq 'HASH' or croak "Invalid $k (not a hash reference)";
-                $self->$method($_ => $$v{$_}) for keys %$v;
+                while (my($name, $code) = each %$v) {
+                    $name =~ /^\w+$/
+                        or croak "Invalid filter name '$name'";
+                    if (defined $code) {
+                        ref $code eq 'CODE'
+                            or $code = ($code =~ /(.*)::(.*)/
+                            ? $1->can($2)
+                            : do {
+                                my($i,$p) = 0;
+                                $i++ while __PACKAGE__ eq ($p = caller $i);
+                                $p->can($code)
+                            })
+                            or croak "Invalid subroutine '$$v{$name}' for filter '$name'";
+                        $$self{filters}{$name} = $code;
+                    } else {
+                        delete $$self{filters}{$name};
+                    }
+                }
+            } else {
+                undef %{$$self{$k}};
+            }
+            next;
+        } elsif ($k eq 'vars') {
+            if (defined $v) {
+                ref $v eq 'HASH' or croak "Invalid $k (not a hash reference)";
+                while (my($name, $ref) = each %$v) {
+                    if (defined $ref) {
+                        my $sigil = $sigil{Scalar::Util::reftype $ref // 'CODE'} // '$';
+                        $name =~ s/^\Q$sigil\E?/$sigil ne '&' && $sigil/e;
+                        $$self{vars}{$name} = $ref;
+                    } else {
+                        delete $$self{vars}{$name};
+                    }
+                }
             } else {
                 undef %{$$self{$k}};
             }
